/*      -*- OpenSAF  -*-
 *
 * (C) Copyright 2008 The OpenSAF Foundation
 * Copyright (C) 2017, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. This file and program are licensed
 * under the GNU Lesser General Public License Version 2.1, February 1999.
 * The complete license can be accessed from the following location:
 * http://opensource.org/licenses/lgpl-license.php
 * See the Copying file included with the OpenSAF distribution for full
 * licensing terms.
 *
 * Author(s): Emerson Network Power
 *
 */

/*****************************************************************************
..............................................................................

..............................................................................

  DESCRIPTION:

  This module contains routines to create, modify, delete & fetch the SU-SI
  and component CSI relationship records.

..............................................................................

  FUNCTIONS INCLUDED in this module:


******************************************************************************
*/

#include "amf/amfnd/avnd.h"
#include "osaf/immutil/immutil.h"

/* static function declarations */

static uint32_t avnd_su_si_csi_rec_modify(AVND_CB *, AVND_SU *,
                                          AVND_SU_SI_REC *,
                                          AVND_COMP_CSI_PARAM *);

static uint32_t avnd_su_si_csi_all_modify(AVND_CB *, AVND_SU *,
                                          AVND_COMP_CSI_PARAM *);

static uint32_t avnd_su_si_csi_rec_del(AVND_CB *, AVND_SU *, AVND_SU_SI_REC *,
                                       AVND_COMP_CSI_REC *);

static uint32_t avnd_su_si_csi_del(AVND_CB *, AVND_SU *, AVND_SU_SI_REC *);

/* macro to add a csi-record to the si-csi list */
#define m_AVND_SU_SI_CSI_REC_ADD(si, csi, rc)                      \
  {                                                                \
    (csi).si_dll_node.key = (uint8_t *)&(csi).rank;                \
    rc = ncs_db_link_list_add(&(si).csi_list, &(csi).si_dll_node); \
  };

/* macro to add a susi record to the beginning of the susi queue */
#define m_AVND_SUDB_REC_SIQ_ADD(su, susi, rc) \
  (rc) = ncs_db_link_list_add(&(su).siq, &(susi).su_dll_node);

/**
 * Initialize the SI list
 * @param cb
 */
void avnd_silist_init(AVND_CB *cb) {
  cb->si_list.order = NCS_DBLIST_ASSCEND_ORDER;
  cb->si_list.cmp_cookie = avsv_dblist_uns32_cmp;
  cb->si_list.free_cookie = 0;
}

/**
 * Get first SI from SI list
 *
 * @return AVND_SU_SI_REC*
 */
AVND_SU_SI_REC *avnd_silist_getfirst(void) {
  NCS_DB_LINK_LIST_NODE *p = m_NCS_DBLIST_FIND_FIRST(&avnd_cb->si_list);

  if (p != nullptr)
    return (
        AVND_SU_SI_REC *)((char *)p -
                          ((char *)&((AVND_SU_SI_REC *)nullptr)->cb_dll_node));
  else
    return nullptr;
}

/**
 * Get next SI from SI list
 * @param si if nullptr first SI is returned
 *
 * @return AVND_SU_SI_REC*
 */
AVND_SU_SI_REC *avnd_silist_getnext(const AVND_SU_SI_REC *si) {
  if (si) {
    NCS_DB_LINK_LIST_NODE *p = m_NCS_DBLIST_FIND_NEXT(&si->cb_dll_node);

    if (p != nullptr)
      return (AVND_SU_SI_REC *)((char *)p -
                                ((char *)&((AVND_SU_SI_REC *)nullptr)
                                     ->cb_dll_node));
    else
      return nullptr;
  } else
    return avnd_silist_getfirst();
}

/**
 * Get previous SI from SI list
 * @param si
 *
 * @return AVND_SU_SI_REC*
 */
AVND_SU_SI_REC *avnd_silist_getprev(const AVND_SU_SI_REC *si) {
  NCS_DB_LINK_LIST_NODE *p = m_NCS_DBLIST_FIND_PREV(&si->cb_dll_node);

  if (p != nullptr)
    return (
        AVND_SU_SI_REC *)((char *)p -
                          ((char *)&((AVND_SU_SI_REC *)nullptr)->cb_dll_node));
  else
    return nullptr;
}

/**
 * Get last SI from SI list
 *
 * @return AVND_SU_SI_REC*
 */

AVND_SU_SI_REC *avnd_silist_getlast(void) {
  NCS_DB_LINK_LIST_NODE *p = m_NCS_DBLIST_FIND_LAST(&avnd_cb->si_list);

  if (p != nullptr)
    return (
        AVND_SU_SI_REC *)((char *)p -
                          ((char *)&((AVND_SU_SI_REC *)nullptr)->cb_dll_node));
  else
    return nullptr;
}

/****************************************************************************
  Name          : avnd_su_si_rec_add

  Description   : This routine creates an su-si relationship record. It
                  updates the su-si record & creates the records in the
                  comp-csi list. If an su-si relationship record already
                  exists, nothing is done.

  Arguments     : cb    - ptr to AvND control block
                  su    - ptr to the AvND SU
                  param - ptr to the SI parameters
                  rc    - ptr to the operation result

  Return Values : ptr to the su-si relationship record

  Notes         : None
******************************************************************************/
AVND_SU_SI_REC *avnd_su_si_rec_add(AVND_CB *cb, AVND_SU *su,
                                   AVND_SU_SI_PARAM *param, uint32_t *rc) {
  AVND_SU_SI_REC *si_rec = 0;
  AVND_COMP_CSI_PARAM *csi_param = 0;

  TRACE_ENTER();

  *rc = NCSCC_RC_SUCCESS;

  /* verify if su-si relationship already exists */
  if (0 != avnd_su_si_rec_get(cb, Amf::to_string(&param->su_name),
                              Amf::to_string(&param->si_name))) {
    *rc = AVND_ERR_DUP_SI;
    goto err;
  }

  /* a fresh si... */
  si_rec = new AVND_SU_SI_REC();

  /*
   * Update the supplied parameters.
   */
  /* update the si-name (key) */
  si_rec->name = Amf::to_string(&param->si_name);
  si_rec->rank = param->si_rank;
  si_rec->curr_state = param->ha_state;

  /*
   * Update the rest of the parameters with default values.
   */
  TRACE("Marking curr and previous assigned state of '%s' unassigned.",
        si_rec->name.c_str());
  m_AVND_SU_SI_CURR_ASSIGN_STATE_SET(si_rec,
                                     AVND_SU_SI_ASSIGN_STATE_UNASSIGNED);
  m_AVND_SU_SI_PRV_ASSIGN_STATE_SET(si_rec, AVND_SU_SI_ASSIGN_STATE_UNASSIGNED);

  /*
   * Add the csi records.
   */
  /* initialize the csi-list (maintained by si) */
  si_rec->csi_list.order = NCS_DBLIST_ASSCEND_ORDER;
  si_rec->csi_list.cmp_cookie = avsv_dblist_uns32_cmp;
  si_rec->csi_list.free_cookie = 0;

  /*
   * Add to the si-list (maintained by su)
   */
  m_AVND_SUDB_REC_SI_ADD(*su, *si_rec, *rc);
  if (NCSCC_RC_SUCCESS != *rc) {
    *rc = AVND_ERR_DLL;
    goto err;
  }

  /* Add to global SI list sorted by rank if appl SU */
  if (!su->is_ncs) {
    uint32_t res;
    si_rec->cb_dll_node.key = (uint8_t *)&si_rec->rank;
    res = ncs_db_link_list_add(&cb->si_list, &si_rec->cb_dll_node);
    osafassert(res == NCSCC_RC_SUCCESS);
  }

  /*
   * Update links to other entities.
   */
  si_rec->su = su;
  si_rec->su_name = su->name;

  /* now add the csi records */
  csi_param = param->list;
  while (0 != csi_param) {
    avnd_su_si_csi_rec_add(cb, su, si_rec, csi_param, rc);
    if (NCSCC_RC_SUCCESS != *rc) goto err;
    csi_param = csi_param->next;
  }

  TRACE_1("SU-SI record added, '%s', '%s', rank:%u",
          osaf_extended_name_borrow(&param->su_name),
          osaf_extended_name_borrow(&param->si_name), si_rec->rank);
  return si_rec;

err:
  if (si_rec) {
    avnd_su_si_csi_del(cb, su, si_rec);
    delete si_rec;
  }

  LOG_CR("SU-SI record addition failed, SU= %s : SI=%s",
         osaf_extended_name_borrow(&param->su_name),
         osaf_extended_name_borrow(&param->si_name));
  TRACE_LEAVE();
  return 0;
}

static void get_cstype(SaImmHandleT immOmHandle,
                       SaImmAccessorHandleT accessorHandle,
                       const std::string &csi_name, std::string &cstype) {
  SaAisErrorT error;
  const SaImmAttrValuesT_2 **attributes;
  SaImmAttrNameT attributeNames[2] = {const_cast<SaImmAttrNameT>("saAmfCSType"),
                                      nullptr};
  const char *type;

  // TODO remove, just for test
  LOG_NO("get_cstype: csi = '%s'", csi_name.c_str());

  if ((error = amf_saImmOmAccessorGet_o2(
           immOmHandle, accessorHandle, csi_name, attributeNames,
           (SaImmAttrValuesT_2 ***)&attributes)) != SA_AIS_OK) {
    LOG_ER("amf_saImmOmAccessorGet_o2 FAILED %u for %s", error,
           csi_name.c_str());
    osafassert(0);
  }

  if ((type = immutil_getStringAttr(attributes, "saAmfCSType", 0)) == nullptr)
    osafassert(0);
  cstype = type;
}

static SaAmfCompCapabilityModelT get_comp_capability(
    const std::string &comp_type, const std::string &csi_name) {
  SaAisErrorT error;
  SaNameT dn = {};
  SaImmAccessorHandleT accessorHandle = 0;
  const SaImmAttrValuesT_2 **attributes;
  SaAmfCompCapabilityModelT comp_cap{};
  SaImmAttrNameT attributeNames[2] = {
      const_cast<SaImmAttrNameT>("saAmfCtCompCapability"), nullptr};
  SaImmHandleT immOmHandle = 0;
  SaVersionT immVersion = {'A', 2, 15};
  std::string cs_type;
  SaNameT comp_type_sanamet;
  SaNameT cs_type_sanamet;

  // TODO remove, just for test
  LOG_NO("get_comp_capability: comptype = '%s' : csi = '%s'", comp_type.c_str(),
         csi_name.c_str());

  TRACE_ENTER2("comptype = '%s' : csi = '%s'", comp_type.c_str(),
               csi_name.c_str());

  error = saImmOmInitialize_cond(&immOmHandle, nullptr, &immVersion);
  if (error != SA_AIS_OK) {
    // TODO - what should comp_cap be?
    LOG_CR("saImmOmInitialize failed: %u", error);
    goto done1;
  }
  amf_saImmOmAccessorInitialize(immOmHandle, accessorHandle);

  get_cstype(immOmHandle, accessorHandle, csi_name, cs_type);
  osaf_extended_name_lend(comp_type.c_str(), &comp_type_sanamet);
  osaf_extended_name_lend(cs_type.c_str(), &cs_type_sanamet);
  avsv_create_association_class_dn(&cs_type_sanamet, &comp_type_sanamet,
                                   "safSupportedCsType", &dn);

  if ((error = amf_saImmOmAccessorGet_o2(
           immOmHandle, accessorHandle, osaf_extended_name_borrow(&dn),
           attributeNames, (SaImmAttrValuesT_2 ***)&attributes)) != SA_AIS_OK) {
    LOG_ER("amf_saImmOmAccessorGet_o2 FAILED %u for'%s'", error,
           osaf_extended_name_borrow(&dn));
    goto done;
  }

  if (immutil_getAttr(const_cast<SaImmAttrNameT>("saAmfCtCompCapability"),
                      attributes, 0, &comp_cap) != SA_AIS_OK)
    osafassert(0);

done:
  immutil_saImmOmAccessorFinalize(accessorHandle);
  immutil_saImmOmFinalize(immOmHandle);
done1:
  osaf_extended_name_free(&dn);
  TRACE_LEAVE2("%u", comp_cap);
  return comp_cap;
}

/****************************************************************************
  Name          : avnd_su_si_csi_rec_add

  Description   : This routine creates a comp-csi relationship record & adds
                  it to the 2 csi lists (maintained by si & comp).

  Arguments     : cb     - ptr to AvND control block
                  su     - ptr to the AvND SU
                  si_rec - ptr to the SI record
                  param  - ptr to the CSI parameters
                  rc     - ptr to the operation result

  Return Values : ptr to the comp-csi relationship record

  Notes         : None
******************************************************************************/
AVND_COMP_CSI_REC *avnd_su_si_csi_rec_add(AVND_CB *cb, AVND_SU *su,
                                          AVND_SU_SI_REC *si_rec,
                                          AVND_COMP_CSI_PARAM *param,
                                          uint32_t *rc) {
  AVND_COMP_CSI_REC *csi_rec = nullptr;
  AVND_COMP *comp = nullptr;
  const std::string param_csi_name = Amf::to_string(&param->csi_name);
  const std::string param_comp_name = Amf::to_string(&param->comp_name);

  TRACE_ENTER2("Comp'%s', Csi'%s' and Rank'%u'", param_comp_name.c_str(),
               param_csi_name.c_str(), param->csi_rank);

  *rc = NCSCC_RC_SUCCESS;

  /* verify if csi record already exists */
  if (0 != avnd_compdb_csi_rec_get(cb, param_comp_name, param_csi_name)) {
    TRACE("csi rec get Failed from compdb");
    *rc = AVND_ERR_DUP_CSI;
    goto err;
  }

  /* get the comp */
  comp = avnd_compdb_rec_get(cb->compdb, param_comp_name);
  if (!comp) {
    /* This could be because of NPI components, NPI components are not added in
       to DB because amfd doesn't send SU presence message to amfnd when SU is
       unlock-in. So, add the component into DB now. */
    if (avnd_comp_config_get_su(su) != NCSCC_RC_SUCCESS) {
      m_AVND_SU_REG_FAILED_SET(su);
      /* Will transition to instantiation-failed when instantiated */
      LOG_ER("su comp config get failed for NPI component:%s",
             param_comp_name.c_str());
      *rc = AVND_ERR_NO_SU;
      goto err;
    }
    comp = avnd_compdb_rec_get(cb->compdb, param_comp_name);
    if (!comp) {
      LOG_ER("comp rec get failed component:%s", param_comp_name.c_str());
      *rc = AVND_ERR_NO_COMP;
      osafassert(0);
      goto err;
    }
  }

  /* a fresh csi... */
  csi_rec = new AVND_COMP_CSI_REC();

  /*
   * Update the supplied parameters.
   */
  /* update the csi-name & csi-rank (keys to comp-csi & si-csi lists resp) */
  csi_rec->name = param_csi_name;
  csi_rec->rank = param->csi_rank;

  /* If CSI capability is not valid, read it from IMM */
  if (param->capability == ~0)
    csi_rec->capability =
        get_comp_capability(comp->saAmfCompType, csi_rec->name);
  else
    csi_rec->capability = param->capability;

  /* update the assignment related parameters */
  csi_rec->act_comp_name = Amf::to_string(&param->active_comp_name);
  csi_rec->trans_desc = param->active_comp_dsc;
  csi_rec->standby_rank = param->stdby_rank;

  /* update the csi-attrs.. steal it from param */
  csi_rec->attrs.number = param->attrs.number;
  csi_rec->attrs.list = param->attrs.list;
  param->attrs.number = 0;
  param->attrs.list = 0;

  /*
   * Update the rest of the parameters with default values.
   */
  TRACE("Marking curr assigned state of '%s' unassigned.",
        csi_rec->name.c_str());
  m_AVND_COMP_CSI_CURR_ASSIGN_STATE_SET(csi_rec,
                                        AVND_COMP_CSI_ASSIGN_STATE_UNASSIGNED);
  m_AVND_COMP_CSI_PRV_ASSIGN_STATE_SET(csi_rec,
                                       AVND_COMP_CSI_ASSIGN_STATE_UNASSIGNED);

  /*
   * Add to the csi-list (maintained by si).
   */
  m_AVND_SU_SI_CSI_REC_ADD(*si_rec, *csi_rec, *rc);
  if (NCSCC_RC_SUCCESS != *rc) {
    *rc = AVND_ERR_DLL;
    goto err;
  }

  /*
   * Add to the csi-list (maintained by comp).
   */
  m_AVND_COMPDB_REC_CSI_ADD(*comp, *csi_rec, *rc);
  if (NCSCC_RC_SUCCESS != *rc) {
    *rc = AVND_ERR_DLL;
    goto err;
  }

  /*
   * Update links to other entities.
   */
  csi_rec->si = si_rec;
  csi_rec->comp = comp;
  csi_rec->comp_name = comp->name;
  csi_rec->si_name = si_rec->name;
  csi_rec->su_name = su->name;
  csi_rec->pending_removal = false;
  csi_rec->suspending_assignment = false;
  return csi_rec;

err:
  if (csi_rec) {
    osafassert(comp != nullptr);
    /* remove from comp-csi & si-csi lists */
    ncs_db_link_list_delink(&si_rec->csi_list, &csi_rec->si_dll_node);
    m_AVND_COMPDB_REC_CSI_REM(*comp, *csi_rec);
    delete csi_rec;
  }

  LOG_CR("Comp-CSI record addition failed, Comp=%s : CSI=%s",
         param_comp_name.c_str(), param_csi_name.c_str());
  TRACE_LEAVE();
  return 0;
}

/****************************************************************************
  Name          : avnd_su_si_rec_modify

  Description   : This routine modifies an su-si relationship record. It
                  updates the su-si record & modifies the records in the
                  comp-csi list.

  Arguments     : cb    - ptr to AvND control block
                  su    - ptr to the AvND SU
                  param - ptr to the SI parameters
                  rc    - ptr to the operation result

  Return Values : ptr to the modified su-si relationship record

  Notes         : None
******************************************************************************/
AVND_SU_SI_REC *avnd_su_si_rec_modify(AVND_CB *cb, AVND_SU *su,
                                      AVND_SU_SI_PARAM *param, uint32_t *rc) {
  AVND_SU_SI_REC *si_rec = 0;

  TRACE_ENTER2();

  *rc = NCSCC_RC_SUCCESS;

  /* get the su-si relationship record */
  si_rec = avnd_su_si_rec_get(cb, Amf::to_string(&param->su_name),
                              Amf::to_string(&param->si_name));
  if (!si_rec) {
    *rc = AVND_ERR_NO_SI;
    goto err;
  }

  /* store the prv state & update the new state */
  si_rec->prv_state = si_rec->curr_state;
  si_rec->curr_state = param->ha_state;

  /* store the prv assign-state & update the new assign-state */
  si_rec->prv_assign_state = si_rec->curr_assign_state;
  TRACE_1("Marking curr assigned state of '%s' unassigned.",
          si_rec->name.c_str());
  m_AVND_SU_SI_CURR_ASSIGN_STATE_SET(si_rec,
                                     AVND_SU_SI_ASSIGN_STATE_UNASSIGNED);

  /* now modify the csi records */
  *rc = avnd_su_si_csi_rec_modify(cb, su, si_rec, param->list);
  if (*rc != NCSCC_RC_SUCCESS) goto err;
  TRACE_LEAVE();
  return si_rec;

err:
  TRACE_LEAVE();
  return 0;
}

/****************************************************************************
  Name          : avnd_su_si_csi_rec_modify

  Description   : This routine modifies a comp-csi relationship record.

  Arguments     : cb     - ptr to AvND control block
                  su     - ptr to the AvND SU
                  si_rec - ptr to the SI record
                  param  - ptr to the CSI parameters

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None
******************************************************************************/
uint32_t avnd_su_si_csi_rec_modify(AVND_CB *cb, AVND_SU *su,
                                   AVND_SU_SI_REC *si_rec,
                                   AVND_COMP_CSI_PARAM *csi_param) {
  AVND_COMP_CSI_PARAM *curr_param = 0;
  AVND_COMP_CSI_REC *curr_csi = 0;
  uint32_t rc = NCSCC_RC_SUCCESS;
  AVND_COMP_CSI_PARAM *param((SA_AMF_HA_QUIESCED == si_rec->curr_state ||
                              SA_AMF_HA_QUIESCING == si_rec->curr_state) ?
		                0 : csi_param);
  TRACE_ENTER2("%p", param);
  /* pick up all the csis belonging to the si & modify them */
  if (!param) {
    TRACE_1("Marking curr assigned state of all CSIs of '%s' unassigned.",
            si_rec->name.c_str());
    for (curr_csi =
             (AVND_COMP_CSI_REC *)m_NCS_DBLIST_FIND_FIRST(&si_rec->csi_list);
         curr_csi; curr_csi = (AVND_COMP_CSI_REC *)m_NCS_DBLIST_FIND_NEXT(
                       &curr_csi->si_dll_node)) {
      if (csi_param) {
        curr_csi->act_comp_name = Amf::to_string(&csi_param->active_comp_name);
        curr_csi->trans_desc    = csi_param->active_comp_dsc;
        curr_csi->standby_rank  = csi_param->stdby_rank;
      }

      /* store the prv assign-state & update the new assign-state */
      curr_csi->prv_assign_state = curr_csi->curr_assign_state;
      m_AVND_COMP_CSI_CURR_ASSIGN_STATE_SET(
          curr_csi, AVND_COMP_CSI_ASSIGN_STATE_UNASSIGNED);
    } /* for */
  }

  /* pick up the csis belonging to the comps specified in the param-list */
  for (curr_param = param; curr_param; curr_param = curr_param->next) {
    /* get the comp & csi */
    curr_csi =
        avnd_compdb_csi_rec_get(cb, Amf::to_string(&curr_param->comp_name),
                                Amf::to_string(&curr_param->csi_name));
    if (!curr_csi || (curr_csi->comp->su != su)) {
      rc = NCSCC_RC_FAILURE;
      goto done;
    }

    /* update the assignment related parameters */
    curr_csi->act_comp_name = Amf::to_string(&curr_param->active_comp_name);
    curr_csi->trans_desc = curr_param->active_comp_dsc;
    curr_csi->standby_rank = curr_param->stdby_rank;

    /* store the prv assign-state & update the new assign-state */
    curr_csi->prv_assign_state = curr_csi->curr_assign_state;
    TRACE("Marking curr assigned state of '%s' unassigned.",
          curr_csi->name.c_str());
    m_AVND_COMP_CSI_CURR_ASSIGN_STATE_SET(
        curr_csi, AVND_COMP_CSI_ASSIGN_STATE_UNASSIGNED);
  } /* for */

done:
  TRACE_LEAVE();
  return rc;
}

/****************************************************************************
  Name          : avnd_su_si_all_modify

  Description   : This routine modifies all the SU-SI & comp-csi records in
                  the database.

  Arguments     : cb    - ptr to AvND control block
                  su    - ptr to the AvND SU
                  param - ptr to the SI parameters

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None
******************************************************************************/
uint32_t avnd_su_si_all_modify(AVND_CB *cb, AVND_SU *su,
                               AVND_SU_SI_PARAM *param) {
  AVND_SU_SI_REC *curr_si = 0;
  uint32_t rc = NCSCC_RC_SUCCESS;

  TRACE_ENTER2();
  /* modify all the si records */
  TRACE("Marking curr assigned state all SIs in '%s' unassigned.",
        su->name.c_str());
  for (curr_si = (AVND_SU_SI_REC *)m_NCS_DBLIST_FIND_FIRST(&su->si_list);
       curr_si; curr_si = (AVND_SU_SI_REC *)m_NCS_DBLIST_FIND_NEXT(
                    &curr_si->su_dll_node)) {
    /* store the prv state & update the new state */
    curr_si->prv_state = curr_si->curr_state;
    curr_si->curr_state = param->ha_state;

    /* store the prv assign-state & update the new assign-state */
    curr_si->prv_assign_state = curr_si->curr_assign_state;
    m_AVND_SU_SI_CURR_ASSIGN_STATE_SET(curr_si,
                                       AVND_SU_SI_ASSIGN_STATE_UNASSIGNED);
  } /* for */

  if (su->si_list.n_nodes > 1)
    LOG_NO("Assigning 'all (%u) SIs' %s to '%s'", su->si_list.n_nodes,
           ha_state[param->ha_state], su->name.c_str());

  /* now modify the comp-csi records */
  rc = avnd_su_si_csi_all_modify(cb, su, param->list);

  TRACE_LEAVE();
  return rc;
}

/****************************************************************************
  Name          : avnd_su_si_csi_all_modify

  Description   : This routine modifies the csi records.

  Arguments     : cb     - ptr to AvND control block
                  su     - ptr to the AvND SU
                  param  - ptr to the CSI parameters (if 0, => all the CSIs
                           belonging to all the SIs in the SU are modified.
                           Else all the CSIs belonging to all the components
                           in the SU are modified)

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None
******************************************************************************/
uint32_t avnd_su_si_csi_all_modify(AVND_CB *cb, AVND_SU *su,
                                   AVND_COMP_CSI_PARAM *param) {
  AVND_COMP_CSI_REC *curr_csi = 0;
  AVND_SU_SI_REC *curr_si = 0;
  uint32_t rc = NCSCC_RC_SUCCESS;

  TRACE_ENTER2("%p", param);
  /* pick up all the csis belonging to all the sis & modify them */
  TRACE("Marking curr assigned state all CSIs in SIs of '%s' unassigned.",
          su->name.c_str());
  for (curr_si = (AVND_SU_SI_REC *)m_NCS_DBLIST_FIND_FIRST(&su->si_list);
       curr_si; curr_si = (AVND_SU_SI_REC *)m_NCS_DBLIST_FIND_NEXT(
                    &curr_si->su_dll_node)) {
    for (curr_csi =
             (AVND_COMP_CSI_REC *)m_NCS_DBLIST_FIND_FIRST(&curr_si->csi_list);
         curr_csi; curr_csi = (AVND_COMP_CSI_REC *)m_NCS_DBLIST_FIND_NEXT(
                       &curr_csi->si_dll_node)) {
      if (param) {
        curr_csi->act_comp_name = Amf::to_string(&param->active_comp_name);
        curr_csi->trans_desc    = param->active_comp_dsc;
        curr_csi->standby_rank  = param->stdby_rank;
      }

      /* store the prv assign-state & update the new assign-state */
      curr_csi->prv_assign_state = curr_csi->curr_assign_state;
      m_AVND_COMP_CSI_CURR_ASSIGN_STATE_SET(
          curr_csi, AVND_COMP_CSI_ASSIGN_STATE_UNASSIGNED);
    } /* for */
  }   /* for */

  TRACE_LEAVE();
  return rc;
}

/****************************************************************************
  Name          : avnd_su_si_rec_del

  Description   : This routine deletes a su-si relationship record. It
                  traverses the entire csi-list and deletes each comp-csi
                  relationship record.

  Arguments     : cb          - ptr to AvND control block
                  su_name - ptr to the su-name
                  si_name - ptr to the si-name

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None
******************************************************************************/
uint32_t avnd_su_si_rec_del(AVND_CB *cb, const std::string &su_name,
                            const std::string &si_name) {
  AVND_SU *su = 0;
  AVND_SU_SI_REC *si_rec = 0;
  uint32_t rc = NCSCC_RC_SUCCESS;
  TRACE_ENTER2("'%s' : '%s'", su_name.c_str(), si_name.c_str());

  /* get the su record */
  su = cb->sudb.find(su_name);
  if (!su) {
    rc = AVND_ERR_NO_SU;
    goto err;
  }

  /* get the si record */
  si_rec = avnd_su_si_rec_get(cb, su_name, si_name);
  if (!si_rec) {
    rc = AVND_ERR_NO_SI;
    goto err;
  }

  /*
   * Delete the csi-list.
   */
  rc = avnd_su_si_csi_del(cb, su, si_rec);
  if (NCSCC_RC_SUCCESS != rc) goto err;

  /*
   * Detach from the si-list (maintained by su).
   */
  rc = m_AVND_SUDB_REC_SI_REM(*su, *si_rec);
  if (NCSCC_RC_SUCCESS != rc) goto err;

  /* remove from global SI list */
  (void)ncs_db_link_list_delink(&cb->si_list, &si_rec->cb_dll_node);

  TRACE_1("SU-SI record deleted, SU= %s : SI=%s", su_name.c_str(),
          si_name.c_str());

  /* free the memory */
  delete si_rec;

  return rc;

err:
  LOG_CR("SU-SI record deletion failed, SU= %s : SI=%s", su_name.c_str(),
         si_name.c_str());
  return rc;
}

/****************************************************************************
  Name          : avnd_su_si_del

  Description   : This routine traverses the entire si-list and deletes each
                  record.

  Arguments     : cb          - ptr to AvND control block
                  su_name - ptr to the su-name (n/w order)

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None
******************************************************************************/
uint32_t avnd_su_si_del(AVND_CB *cb, const std::string &su_name) {
  AVND_SU *su = 0;
  AVND_SU_SI_REC *si_rec = 0;
  uint32_t rc = NCSCC_RC_SUCCESS;

  TRACE_ENTER2("'%s'", su_name.c_str());

  /* get the su record */
  su = cb->sudb.find(su_name);
  if (!su) {
    rc = AVND_ERR_NO_SU;
    goto err;
  }

  /* scan & delete each si record */
  while (0 !=
         (si_rec = (AVND_SU_SI_REC *)m_NCS_DBLIST_FIND_FIRST(&su->si_list))) {
    rc = avnd_su_si_rec_del(cb, su_name, si_rec->name);
    if (NCSCC_RC_SUCCESS != rc) goto err;
  }

err:
  TRACE_LEAVE2("%u", rc);
  return rc;
}

/****************************************************************************
  Name          : avnd_su_si_csi_del

  Description   : This routine traverses the each record in the csi-list
                  (maintained by si) & deletes them.

  Arguments     : cb      - ptr to AvND control block
                  su      - ptr to the AvND SU
                  si_rec  - ptr to the SI record

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None
******************************************************************************/
uint32_t avnd_su_si_csi_del(AVND_CB *cb, AVND_SU *su, AVND_SU_SI_REC *si_rec) {
  AVND_COMP_CSI_REC *csi_rec = 0;
  uint32_t rc = NCSCC_RC_SUCCESS;
  TRACE_ENTER2("'%s' : '%s'", su->name.c_str(), si_rec->name.c_str());

  /* scan & delete each csi record */
  while (0 != (csi_rec = (AVND_COMP_CSI_REC *)m_NCS_DBLIST_FIND_FIRST(
                   &si_rec->csi_list))) {
    rc = avnd_su_si_csi_rec_del(cb, si_rec->su, si_rec, csi_rec);
    if (NCSCC_RC_SUCCESS != rc) goto err;
  }
  TRACE_LEAVE2("%u", rc);
  return rc;

err:
  TRACE_LEAVE2("%u", rc);
  return rc;
}

/****************************************************************************
  Name          : avnd_su_si_csi_rec_del

  Description   : This routine deletes a comp-csi relationship record.

  Arguments     : cb      - ptr to AvND control block
                  su      - ptr to the AvND SU
                  si_rec  - ptr to the SI record
                  csi_rec - ptr to the CSI record

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None
******************************************************************************/
uint32_t avnd_su_si_csi_rec_del(AVND_CB *cb, AVND_SU *su,
                                AVND_SU_SI_REC *si_rec,
                                AVND_COMP_CSI_REC *csi_rec) {
  uint16_t i;

  TRACE_ENTER2("'%s' : '%s' : '%s'", su->name.c_str(), si_rec->name.c_str(),
               csi_rec->name.c_str());

  /* remove from the comp-csi list */
  uint32_t rc = m_AVND_COMPDB_REC_CSI_REM(*(csi_rec->comp), *csi_rec);
  if (NCSCC_RC_SUCCESS != rc) goto err;

  /* if all csi's are removed & the pres state is un-instantiated for
   * a  npi comp, its time to mark it as healthy
   */
  if (m_AVND_COMP_IS_FAILED(csi_rec->comp) &&
      !m_AVND_COMP_TYPE_IS_PREINSTANTIABLE(csi_rec->comp) &&
      m_AVND_COMP_PRES_STATE_IS_UNINSTANTIATED(csi_rec->comp) &&
      csi_rec->comp->csi_list.n_nodes == 0) {
    m_AVND_COMP_FAILED_RESET(csi_rec->comp);
  }

  /* remove from the si-csi list */
  rc = m_AVND_SU_SI_CSI_REC_REM(*si_rec, *csi_rec);
  if (NCSCC_RC_SUCCESS != rc) goto err;

  /*
   * Free the memory alloced for this record.
   */
  // free the csi attributes
  // use of free() is required as it was
  // malloc'ed (eg. in avsv_edp_susi_asgn())
  // Try to free the attributes, in theory, they contain SaNameT
  // So, it is potential for a leak
  for (i = 0; i < csi_rec->attrs.number; i++) {
    osaf_extended_name_free(&csi_rec->attrs.list[i].name);
    osaf_extended_name_free(&csi_rec->attrs.list[i].value);
    free(csi_rec->attrs.list[i].string_ptr);
  }
  //
  free(csi_rec->attrs.list);

  /* free the pg list TBD */
  TRACE_1("Comp-CSI record deletion success, Comp=%s : CSI=%s",
          csi_rec->comp->name.c_str(), csi_rec->name.c_str());

  /* finally free this record */
  delete csi_rec;

  TRACE_LEAVE();
  return rc;

err:
  LOG_CR("Comp-CSI record deletion failed, Comp=%s : CSI=%s",
         csi_rec->comp->name.c_str(), csi_rec->name.c_str());
  return rc;
}

/****************************************************************************
  Name          : avnd_su_si_rec_get

  Description   : This routine gets the su-si relationship record from the
                  si-list (maintained on su).

  Arguments     : cb          - ptr to AvND control block
                  su_name - ptr to the su-name (n/w order)
                  si_name - ptr to the si-name (n/w order)

  Return Values : ptr to the su-si record (if any)

  Notes         : None
******************************************************************************/
AVND_SU_SI_REC *avnd_su_si_rec_get(AVND_CB *cb, const std::string &su_name,
                                   const std::string &si_name) {
  AVND_SU_SI_REC *si_rec = 0;
  AVND_SU *su = 0;

  TRACE_ENTER2("'%s' : '%s'", su_name.c_str(), si_name.c_str());
  /* get the su record */
  su = cb->sudb.find(su_name);
  if (!su) goto done;

  /* get the si record */
  si_rec = (AVND_SU_SI_REC *)ncs_db_link_list_find(&su->si_list,
                                                   (uint8_t *)si_name.c_str());

done:
  TRACE_LEAVE();
  return si_rec;
}

/****************************************************************************
  Name          : avnd_su_siq_rec_add

  Description   : This routine buffers the susi assign message parameters in
                  the susi queue.

  Arguments     : cb    - ptr to AvND control block
                  su    - ptr to the AvND SU
                  param - ptr to the SI parameters
                  rc    - ptr to the operation result

  Return Values : ptr to the si queue record

  Notes         : None
******************************************************************************/
AVND_SU_SIQ_REC *avnd_su_siq_rec_add(AVND_CB *cb, AVND_SU *su,
                                     AVND_SU_SI_PARAM *param, uint32_t *rc) {
  AVND_SU_SIQ_REC *siq = 0;

  *rc = NCSCC_RC_SUCCESS;
  TRACE_ENTER2("'%s'", su->name.c_str());

  /* alloc the siq rec */
  siq = new AVND_SU_SIQ_REC();

  /* Add to the siq (maintained by su) */
  m_AVND_SUDB_REC_SIQ_ADD(*su, *siq, *rc);
  if (NCSCC_RC_SUCCESS != *rc) {
    *rc = AVND_ERR_DLL;
    goto err;
  }

  /* update the param */
  siq->info = *param;

  /* memory transferred to the siq-rec.. nullify it in param */
  param->list = 0;

  TRACE_LEAVE();
  return siq;

err:
  if (siq) delete siq;

  TRACE_LEAVE();
  return 0;
}

/****************************************************************************
  Name          : avnd_su_siq_rec_del

  Description   : This routine deletes the buffered susi assign message from
                  the susi queue.

  Arguments     : cb  - ptr to AvND control block
                  su  - ptr to the AvND SU
                  siq - ptr to the si queue rec

  Return Values : None.

  Notes         : None.
******************************************************************************/
void avnd_su_siq_rec_del(AVND_CB *cb, AVND_SU *su, AVND_SU_SIQ_REC *siq) {
  AVSV_SUSI_ASGN *curr = 0;
  TRACE_ENTER2("'%s'", su->name.c_str());

  /* delete the comp-csi info */
  while ((curr = siq->info.list) != 0) {
    siq->info.list = curr->next;
    // AVSV_ATTR_NAME_VAL variables
    // are malloc'ed, use free()
    free(curr->attrs.list);

    // use of free() is required as it was
    // malloc'ed in avsv_edp_susi_asgn()
    free(curr);
  }

  /* free the rec */
  delete siq;

  TRACE_LEAVE();
  return;
}
