/*      -*- OpenSAF  -*-
 *
 * (C) Copyright 2008 The OpenSAF Foundation
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. This file and program are licensed
 * under the GNU Lesser General Public License Version 2.1, February 1999.
 * The complete license can be accessed from the following location:
 * http://opensource.org/licenses/lgpl-license.php
 * See the Copying file included with the OpenSAF distribution for full
 * licensing terms.
 *
 * Author(s): Emerson Network Power
 *
 */

/*****************************************************************************
..............................................................................

..............................................................................

  DESCRIPTION:
  This file includes the functions require for MDS interface.

  ..............................................................................

  FUNCTIONS INCLUDED in this module:
  mbcsv_mds_reg        - Register with MDS.
  mbcsv_mds_rcv        - rcv something from MDS lower layer.
  mbcsv_mds_evt        - subscription event entry point
  mbcsv_mds_enc        - MDS encode routine.
  mbcsv_mds_dec        - MDS decode routine.
  mbcsv_mds_cpy        - MDS copy routine.
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
*/

#include "mbcsv.h"

MDS_CLIENT_MSG_FORMAT_VER
MBCSV_wrt_PEER_msg_fmt_array[MBCSV_WRT_PEER_SUBPART_VER_RANGE] = {
    1 /* msg format version for subpart version */
};

/****************************************************************************
  PROCEDURE     : mbcsv_mds_reg

  Description   : This routine registers the MBCSV Service with MDS with the
		  pwe handle passed by MBCSv client.

  Arguments     : pwe_hdl - Handle passed by MBCSv client in OPEN API.

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None.
******************************************************************************/
uint32_t mbcsv_mds_reg(uint32_t pwe_hdl, uint32_t svc_hdl, MBCSV_ANCHOR *anchor,
		       MDS_DEST *vdest)
{
	NCSMDS_INFO svc_to_mds_info;
	MDS_SVC_ID svc_ids_array[1];
	TRACE_ENTER2("pwe_hdl:%u, svc_hdl: %u, anchor:%" PRIu64, pwe_hdl,
		     svc_hdl, *anchor);

	if (NCSCC_RC_SUCCESS != mbcsv_query_mds(pwe_hdl, anchor, vdest)) {
		TRACE_LEAVE2(
		    "Invalid handle passed. pwe_hdl: %u, anchor:%" PRIu64,
		    pwe_hdl, *anchor);
		return NCSCC_RC_FAILURE;
	}

	/* Install mds */
	memset(&svc_to_mds_info, 0, sizeof(NCSMDS_INFO));
	svc_to_mds_info.i_mds_hdl = pwe_hdl;
	svc_to_mds_info.i_svc_id = NCSMDS_SVC_ID_MBCSV;
	svc_to_mds_info.i_op = MDS_INSTALL;
	svc_to_mds_info.info.svc_install.i_yr_svc_hdl = pwe_hdl;
	svc_to_mds_info.info.svc_install.i_install_scope = NCSMDS_SCOPE_NONE;
	svc_to_mds_info.info.svc_install.i_svc_cb = mbcsv_mds_callback;
	svc_to_mds_info.info.svc_install.i_mds_q_ownership = false;
	svc_to_mds_info.info.svc_install.i_mds_svc_pvt_ver =
	    MBCSV_MDS_SUB_PART_VERSION;

	if (ncsmds_api(&svc_to_mds_info) != NCSCC_RC_SUCCESS) {
		TRACE_LEAVE2("MDS install failed. pwe_hdl = %u", pwe_hdl);
		return NCSCC_RC_FAILURE;
	}

	/* MBCSV is subscribing for MBCSv events */
	memset(&svc_to_mds_info, 0, sizeof(NCSMDS_INFO));
	svc_to_mds_info.i_mds_hdl = pwe_hdl;
	svc_to_mds_info.i_svc_id = NCSMDS_SVC_ID_MBCSV;
	svc_to_mds_info.i_op = MDS_RED_SUBSCRIBE;
	svc_to_mds_info.info.svc_subscribe.i_scope = NCSMDS_SCOPE_NONE;
	svc_to_mds_info.info.svc_subscribe.i_num_svcs = 1;
	svc_ids_array[0] = NCSMDS_SVC_ID_MBCSV;
	svc_to_mds_info.info.svc_subscribe.i_svc_ids = svc_ids_array;

	if (ncsmds_api(&svc_to_mds_info) != NCSCC_RC_SUCCESS) {
		mbcsv_mds_unreg(pwe_hdl);
		TRACE_LEAVE2(
		    "MDS subscribe failed. pwe_hdl:%u, anchor:%" PRIu64,
		    pwe_hdl, *anchor);
		return NCSCC_RC_FAILURE;
	}

	TRACE_LEAVE();
	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
  PROCEDURE     : mbcsv_query_mds

  Description   : This routine sends MDS query for getting anchor and VDEST;.

  Arguments     : pwe_hdl - Handle passed by MBCSv client in OPEN API.
		  anchor  - anchor value returned by MDS.
		  vdest   - VDEST value returned by MDS.

  Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE

  Notes         : None.
******************************************************************************/
uint32_t mbcsv_query_mds(uint32_t pwe_hdl, MBCSV_ANCHOR *anchor,
			 MDS_DEST *vdest)
{
	NCSMDS_INFO mds_info;
	memset(&mds_info, 0, sizeof(mds_info));

	mds_info.i_mds_hdl = pwe_hdl;
	mds_info.i_op = MDS_QUERY_PWE;
	mds_info.i_svc_id = NCSMDS_SVC_ID_MBCSV;

	if (ncsmds_api(&mds_info) != NCSCC_RC_SUCCESS) {
		TRACE_LEAVE();
		return NCSCC_RC_FAILURE;
	}

	if ((mds_info.info.query_pwe.o_pwe_id == 0) ||
	    (mds_info.info.query_pwe.o_absolute)) {
		TRACE_LEAVE();
		return NCSCC_RC_FAILURE;
	} else {
		*anchor = mds_info.info.query_pwe.info.virt_info.o_anc;
		memcpy(vdest, &mds_info.info.query_pwe.info.virt_info.o_vdest,
		       sizeof(MDS_DEST));
	}

	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
  PROCEDURE     : mbcsv_mds_unreg

  Description   : This routine unregisters the MBCSV Service from MDS on the
		  MBCSVs Vaddress

  Arguments     :  pwe_hdl - Handle with whic chackpoint instance registered
with MBCSv.

  Return Values : NONE

  Notes         : None.
******************************************************************************/
void mbcsv_mds_unreg(uint32_t pwe_hdl)
{
	NCSMDS_INFO svc_to_mds_info;
	TRACE_ENTER();

	svc_to_mds_info.i_mds_hdl = pwe_hdl;
	svc_to_mds_info.i_svc_id = NCSMDS_SVC_ID_MBCSV;
	svc_to_mds_info.i_op = MDS_UNINSTALL;

	if (NCSCC_RC_SUCCESS != ncsmds_api(&svc_to_mds_info)) {
		TRACE_4("MDS uninstall failed. pwe_hdl:%u", pwe_hdl);
	}

	TRACE_LEAVE();
	return;
}

/****************************************************************************
  PROCEDURE     : mbcsv_mds_send_msg

  Description   : This routine is use for sending the message to the MBCSv peer.

  Arguments     : cb - ptr to the MBCSVcontrol block

  Return Values : NONE

  Notes         : None.
******************************************************************************/
uint32_t mbcsv_mds_send_msg(uint32_t send_type, MBCSV_EVT *msg, CKPT_INST *ckpt,
			    MBCSV_ANCHOR anchor)
{
	uint32_t rc;
	NCSMDS_INFO mds_info;
	TRACE_ENTER2("sending to vdest:%" PRIx64, ckpt->my_vdest);

	memset(&mds_info, 0, sizeof(mds_info));

	mds_info.i_mds_hdl = ckpt->pwe_hdl;
	mds_info.i_svc_id = NCSMDS_SVC_ID_MBCSV;
	mds_info.i_op = MDS_SEND;

	mds_info.info.svc_send.i_msg = msg;
	mds_info.info.svc_send.i_to_svc = NCSMDS_SVC_ID_MBCSV;
	mds_info.info.svc_send.i_sendtype = send_type;
	mds_info.info.svc_send.i_priority = MDS_SEND_PRIORITY_HIGH;

	switch (send_type) {
	case MDS_SENDTYPE_RED: {
		TRACE("send type MDS_SENDTYPE_RED");
		mds_info.info.svc_send.info.red.i_to_anc = anchor;
		mds_info.info.svc_send.info.red.i_to_vdest = ckpt->my_vdest;
	} break;

	case MDS_SENDTYPE_REDRSP: {
		TRACE("send type MDS_SENDTYPE_REDRSP:");
		mds_info.info.svc_send.info.redrsp.i_time_to_wait =
		    MBCSV_SYNC_TIMEOUT;
		mds_info.info.svc_send.info.redrsp.i_to_anc = anchor;
		mds_info.info.svc_send.info.redrsp.i_to_vdest = ckpt->my_vdest;
	} break;

	case MDS_SENDTYPE_RRSP: {
		TRACE("send type MDS_SENDTYPE_RRSP:");
		mds_info.info.svc_send.info.rrsp.i_msg_ctxt = msg->msg_ctxt;
		mds_info.info.svc_send.info.rrsp.i_to_anc = anchor;
		mds_info.info.svc_send.info.rrsp.i_to_dest = ckpt->my_vdest;
	} break;

	case MDS_SENDTYPE_RBCAST: {
		TRACE("send type MDS_SENDTYPE_RBCAST");
		m_NCS_MBCSV_MDS_BCAST_SEND(ckpt->pwe_hdl, msg, ckpt);
		TRACE_LEAVE2("success");
		return NCSCC_RC_SUCCESS;
	} break;

	default:
		TRACE_LEAVE2("send type not supported by MBCSV");
		return NCSCC_RC_FAILURE;
	}

	if ((rc = ncsmds_api(&mds_info)) == NCSCC_RC_SUCCESS) {
		/* If message is send resp  then free the message received in
		 * response  */
		if ((MDS_SENDTYPE_REDRSP == send_type) &&
		    (NULL != mds_info.info.svc_send.info.redrsp.o_rsp)) {
			m_MMGR_FREE_MBCSV_EVT(
			    mds_info.info.svc_send.info.redrsp.o_rsp);
		}
		TRACE_LEAVE2("success");
		return NCSCC_RC_SUCCESS;
	} else {
		TRACE_LEAVE2("failure");
		return rc;
	}
}

/****************************************************************************
 * PROCEDURE     : mbcsv_mds_callback
 *
 * Description   : Call back function provided to MDS. MDS will call this
 *                 function for enc/dec/cpy/rcv/svc_evt operations.
 *
 * Arguments     : NCSMDS_CALLBACK_INFO *info: call back info.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.
 *
 * Notes         : None.
 *****************************************************************************/
uint32_t mbcsv_mds_callback(NCSMDS_CALLBACK_INFO *cbinfo)
{
	uint32_t status;

	switch (cbinfo->i_op) {
	case MDS_CALLBACK_COPY:
		status = mbcsv_mds_cpy(
		    cbinfo->i_yr_svc_hdl, cbinfo->info.cpy.i_msg,
		    cbinfo->info.cpy.i_to_svc_id, &cbinfo->info.cpy.o_cpy,
		    cbinfo->info.cpy.i_last, cbinfo->info.cpy.i_rem_svc_pvt_ver,
		    &cbinfo->info.cpy.o_msg_fmt_ver);
		break;

	case MDS_CALLBACK_ENC:
		/* Treating both encode types as same */
		status = mbcsv_mds_enc(
		    cbinfo->i_yr_svc_hdl, cbinfo->info.enc.i_msg,
		    cbinfo->info.enc.i_to_svc_id, cbinfo->info.enc.io_uba,
		    cbinfo->info.enc.i_rem_svc_pvt_ver,
		    &cbinfo->info.enc.o_msg_fmt_ver);
		break;

	case MDS_CALLBACK_ENC_FLAT:
		/* Treating both encode types as same */
		status = mbcsv_mds_enc(cbinfo->i_yr_svc_hdl,
				       cbinfo->info.enc_flat.i_msg,
				       cbinfo->info.enc_flat.i_to_svc_id,
				       cbinfo->info.enc_flat.io_uba,
				       cbinfo->info.enc_flat.i_rem_svc_pvt_ver,
				       &cbinfo->info.enc_flat.o_msg_fmt_ver);
		break;

	case MDS_CALLBACK_DEC:
		/* Treating both decode types as same */
		status = mbcsv_mds_dec(
		    cbinfo->i_yr_svc_hdl, &cbinfo->info.dec.o_msg,
		    cbinfo->info.dec.i_fr_svc_id, cbinfo->info.dec.io_uba,
		    cbinfo->info.dec.i_msg_fmt_ver);
		break;

	case MDS_CALLBACK_DEC_FLAT:
		/* Treating both decode types as same */
		status = mbcsv_mds_dec(cbinfo->i_yr_svc_hdl,
				       &cbinfo->info.dec_flat.o_msg,
				       cbinfo->info.dec_flat.i_fr_svc_id,
				       cbinfo->info.dec_flat.io_uba,
				       cbinfo->info.dec_flat.i_msg_fmt_ver);
		break;

	case MDS_CALLBACK_RECEIVE:
		status = mbcsv_mds_rcv(cbinfo);
		break;

	case MDS_CALLBACK_SVC_EVENT:
		status =
		    mbcsv_mds_evt(cbinfo->info.svc_evt, cbinfo->i_yr_svc_hdl);
		break;

	/* Fix for  MBCSv doesn't handle MDS Quiesced Ack callback */
	case MDS_CALLBACK_DIRECT_RECEIVE:
	case MDS_CALLBACK_QUIESCED_ACK:
		status = NCSCC_RC_SUCCESS;
		break;

	default:
		TRACE_4("Incorrect operation type");
		status = NCSCC_RC_FAILURE;
		break;
	}

	return status;
}

/****************************************************************************
 * Function Name: mbcsv_mds_rcv
 *
 * Description   : MBCSV receive callback function. On receiving message from
 *                 peer, MBCSv will post this message to the MBCSv mailbox for
 *                 further processing.
 *
 * Arguments     : NCSMDS_CALLBACK_INFO *info: call back info.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.
 *
 * Notes         : None.
 *****************************************************************************/

uint32_t mbcsv_mds_rcv(NCSMDS_CALLBACK_INFO *cbinfo)
{
	MBCSV_EVT *msg = (MBCSV_EVT *)cbinfo->info.receive.i_msg;
	SYSF_MBX mbx;
	uint32_t send_pri;

	if (NULL == msg) {
		TRACE_LEAVE2("null message received");
		return NCSCC_RC_FAILURE;
	}

	msg->rcvr_peer_key.pwe_hdl = (uint32_t)cbinfo->i_yr_svc_hdl;
	msg->rcvr_peer_key.peer_anchor = cbinfo->info.receive.i_fr_anc;

	/* Is used during sync send only but will be copied always. */
	msg->msg_ctxt = cbinfo->info.receive.i_msg_ctxt;

	if (0 != (mbx = mbcsv_get_mbx(msg->rcvr_peer_key.pwe_hdl,
				      msg->rcvr_peer_key.svc_id))) {
		/*
		 * We found out the mailbox to which we can post a message. Now
		 * construct and send this message to the mailbox.
		 */
		msg->msg_type = MBCSV_EVT_INTERNAL_RCV;

		if (msg->info.peer_msg.type == MBCSV_EVT_INTERNAL_PEER_DISC) {
			send_pri = NCS_IPC_PRIORITY_HIGH;
		} else
			send_pri = NCS_IPC_PRIORITY_NORMAL;

		if (NCSCC_RC_SUCCESS != m_MBCSV_SND_MSG(&mbx, msg, send_pri)) {
			if (msg->info.peer_msg.type == MBCSV_EVT_INTERNAL_CLIENT) {
				m_MMGR_FREE_BUFR_LIST(
				    msg->info.peer_msg.info.client_msg.uba.ub);
			}
			m_MMGR_FREE_MBCSV_EVT(msg);
			TRACE_LEAVE2("ipc send failed");
			return NCSCC_RC_FAILURE;
		}
	} else {
		if (msg->info.peer_msg.type == MBCSV_EVT_INTERNAL_CLIENT) {
			m_MMGR_FREE_BUFR_LIST(
			    msg->info.peer_msg.info.client_msg.uba.ub);
		}
		m_MMGR_FREE_MBCSV_EVT(msg);
	}

	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
 * Function Name: mbcsv_mds_evt
 *
 * Description  : MBCSV is informed when MDS events occurr that he has
 *                subscribed to. First MBCSv will post this evet to all the
 *                the services which are currently registered on this VDEST.
 *                Later on this event will get processed in the dispatch.
 *
 * Arguments     : NCSMDS_CALLBACK_INFO *info: call back info.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.
 *
 * Notes         : None.
 *****************************************************************************/
uint32_t mbcsv_mds_evt(MDS_CALLBACK_SVC_EVENT_INFO svc_info,
		       MDS_CLIENT_HDL yr_svc_hdl)
{
	MBCSV_EVT *evt;
	SYSF_MBX mbx;
	SS_SVC_ID svc_id = 0;
	MDS_DEST vdest;
	MBCSV_ANCHOR anchor;

	if ((svc_info.i_change != NCSMDS_RED_UP) &&
	    (svc_info.i_change != NCSMDS_RED_DOWN)) {
		return NCSCC_RC_SUCCESS;
	}
	/*
	 * First find out whether MBCSv is registered on this VDEST
	 * If no then return success.
	 */
	if (NCSCC_RC_SUCCESS !=
	    mbcsv_query_mds((uint32_t)svc_info.svc_pwe_hdl, &anchor, &vdest)) {
		TRACE_LEAVE2("mbcsv is not registered on this vdest");
		return NCSCC_RC_FAILURE;
	}

	/* If VDEST not equal, then discard message.(message from other VDESTs)
	 */
	if (0 != memcmp(&vdest, &svc_info.i_dest, sizeof(MDS_DEST))) {
		TRACE_LEAVE2("Msg is not from same vdest, discarding");
		return NCSCC_RC_SUCCESS;
	}

	/* VDEST is my VDEST. So next, check for self events */
	if (anchor == svc_info.i_anc) {
		TRACE_LEAVE2("vdest is same as my vdest");
		return NCSCC_RC_SUCCESS;
	}

	if (svc_info.i_change == NCSMDS_RED_UP) {
		TRACE_1("RED_UP event. pwe_hdl:%u, anchor:%" PRIu64,
			svc_info.svc_pwe_hdl, svc_info.i_anc);

		mbcsv_add_new_pwe_anc((uint32_t)svc_info.svc_pwe_hdl,
				      svc_info.i_anc);
	} else {
		TRACE_1("RED_DOWN event. pwe_hdl: %u, anchor:%" PRIu64,
			svc_info.svc_pwe_hdl, svc_info.i_anc);

		mbcsv_rmv_pwe_anc_entry((uint32_t)svc_info.svc_pwe_hdl,
					svc_info.i_anc);
	}

	while (0 != (mbx = mbcsv_get_next_entry_for_pwe(
			 (uint32_t)svc_info.svc_pwe_hdl, &svc_id))) {
		if (NULL == (evt = m_MMGR_ALLOC_MBCSV_EVT)) {
			TRACE_LEAVE2("malloc failed");
			return NCSCC_RC_FAILURE;
		}

		memset(evt, '\0', sizeof(MBCSV_EVT));

		evt->msg_type = MBCSV_EVT_MDS_SUBSCR;

		evt->rcvr_peer_key.svc_id = svc_id;
		evt->rcvr_peer_key.pwe_hdl = (uint32_t)svc_info.svc_pwe_hdl;
		evt->rcvr_peer_key.peer_anchor = svc_info.i_anc;

		evt->info.mds_sub_evt.evt_type = svc_info.i_change;

		if (NCSCC_RC_SUCCESS !=
		    m_MBCSV_SND_MSG(&mbx, evt, NCS_IPC_PRIORITY_VERY_HIGH)) {
			m_MMGR_FREE_MBCSV_EVT(evt);
			TRACE_LEAVE2("ipc send failed");
			return NCSCC_RC_FAILURE;
		}
	}

	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
 * Function Name: mbcsv_mds_enc
 *
 * Description  : Encode message to tbe sent to the MBCSv peer. We are just
 *                the internal messages which we exchange between the peers.
 *
 * Arguments     : NCSMDS_CALLBACK_INFO *info: call back info.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.
 *
 * Notes         : None.
 *****************************************************************************/
uint32_t mbcsv_mds_enc(MDS_CLIENT_HDL yr_svc_hdl, NCSCONTEXT msg,
		       SS_SVC_ID to_svc, NCS_UBAID *uba,
		       MDS_SVC_PVT_SUB_PART_VER rem_svc_pvt_ver,
		       MDS_CLIENT_MSG_FORMAT_VER *msg_fmt_ver)
{
	uint8_t *data;
	MBCSV_EVT *mm;
	MDS_CLIENT_MSG_FORMAT_VER msg_fmt_version;

	if (uba == NULL) {
		TRACE_LEAVE2("userbuff is null");
		return NCSCC_RC_FAILURE;
	}

	/* Set the current message format version */
	msg_fmt_version = m_NCS_ENC_MSG_FMT_GET(
	    rem_svc_pvt_ver, MBCSV_WRT_PEER_SUBPART_VER_MIN,
	    MBCSV_WRT_PEER_SUBPART_VER_MAX, MBCSV_wrt_PEER_msg_fmt_array);
	if (0 == msg_fmt_version) {
		char str[200];
		snprintf(
		    str, sizeof(str),
		    "Peer MDS Subpart version:%d not supported, message to svc-id:%d dropped",
		    rem_svc_pvt_ver, to_svc);
		TRACE_LEAVE2("%s", str);
		return NCSCC_RC_FAILURE;
	}
	*msg_fmt_ver = msg_fmt_version;

	mm = (MBCSV_EVT *)msg;

	if (mm == NULL) {
		TRACE_LEAVE2("message to be encoded is NULL");
		return NCSCC_RC_FAILURE;
	}

	data = ncs_enc_reserve_space(uba, MBCSV_MSG_TYPE_SIZE);
	if (data == NULL) {
		TRACE_LEAVE2("allocating uba failed");
		return NCSCC_RC_FAILURE;
	}

	ncs_encode_8bit(&data, mm->info.peer_msg.type);
	ncs_encode_32bit(&data, mm->rcvr_peer_key.svc_id);

	ncs_enc_claim_space(uba, MBCSV_MSG_TYPE_SIZE);

	switch (mm->info.peer_msg.type) {
	case MBCSV_EVT_INTERNAL_PEER_DISC: {
		data = ncs_enc_reserve_space(uba, MBCSV_MSG_SUB_TYPE);
		if (data == NULL) {
			TRACE_LEAVE2("allocating uba failed");
			return NCSCC_RC_FAILURE;
		}

		ncs_encode_8bit(&data,
				mm->info.peer_msg.info.peer_disc.msg_sub_type);
		ncs_enc_claim_space(uba, MBCSV_MSG_SUB_TYPE);

		switch (mm->info.peer_msg.info.peer_disc.msg_sub_type) {
		case MBCSV_PEER_UP_MSG: {
			data =
			    ncs_enc_reserve_space(uba, MBCSV_PEER_UP_MSG_SIZE);
			if (data == NULL) {
				TRACE_LEAVE2("allocating uba failed");
				return NCSCC_RC_FAILURE;
			}
			ncs_encode_32bit(
			    &data, mm->info.peer_msg.info.peer_disc.peer_role);
			ncs_enc_claim_space(uba, MBCSV_PEER_UP_MSG_SIZE);

			mbcsv_encode_version(uba,
					     mm->info.peer_msg.info.peer_disc
						 .info.peer_up.peer_version);

			break;
		}

		case MBCSV_PEER_DOWN_MSG: {
			data = ncs_enc_reserve_space(uba,
						     MBCSV_PEER_DOWN_MSG_SIZE);
			if (data == NULL) {
				TRACE_LEAVE2("allocating uba failed");
				return NCSCC_RC_FAILURE;
			}

			ncs_encode_32bit(
			    &data, (uint32_t)mm->rcvr_peer_key.peer_inst_hdl);
			ncs_encode_32bit(
			    &data, mm->info.peer_msg.info.peer_disc.peer_role);
			ncs_enc_claim_space(uba, MBCSV_PEER_DOWN_MSG_SIZE);

			break;
		}

		case MBCSV_PEER_INFO_MSG: {
			data = ncs_enc_reserve_space(uba,
						     MBCSV_PEER_INFO_MSG_SIZE);
			if (data == NULL) {
				TRACE_LEAVE2("allocating uba failed");
				return NCSCC_RC_FAILURE;
			}
			ncs_encode_32bit(&data,
					 mm->info.peer_msg.info.peer_disc.info
					     .peer_info.my_peer_inst_hdl);
			ncs_encode_32bit(
			    &data, mm->info.peer_msg.info.peer_disc.peer_role);
			ncs_encode_8bit(&data, mm->info.peer_msg.info.peer_disc
						   .info.peer_info.compatible);
			ncs_enc_claim_space(uba, MBCSV_PEER_INFO_MSG_SIZE);

			mbcsv_encode_version(uba,
					     mm->info.peer_msg.info.peer_disc
						 .info.peer_up.peer_version);

			break;
		}

		case MBCSV_PEER_INFO_RSP_MSG: {
			data = ncs_enc_reserve_space(
			    uba, MBCSV_PEER_INFO_RSP_MSG_SIZE);
			if (data == NULL) {
				TRACE_LEAVE2("allocating uba failed");
				return NCSCC_RC_FAILURE;
			}

			ncs_encode_32bit(
			    &data, (uint32_t)mm->rcvr_peer_key.peer_inst_hdl);
			ncs_encode_32bit(&data,
					 mm->info.peer_msg.info.peer_disc.info
					     .peer_info_rsp.my_peer_inst_hdl);
			ncs_encode_32bit(
			    &data, mm->info.peer_msg.info.peer_disc.peer_role);
			ncs_encode_8bit(&data,
					mm->info.peer_msg.info.peer_disc.info
					    .peer_info_rsp.compatible);
			ncs_enc_claim_space(uba, MBCSV_PEER_INFO_RSP_MSG_SIZE);

			mbcsv_encode_version(
			    uba, mm->info.peer_msg.info.peer_disc.info
				     .peer_info_rsp.peer_version);

			break;
		}

		case MBCSV_PEER_CHG_ROLE_MSG: {
			data = ncs_enc_reserve_space(
			    uba, MBCSV_PEER_CHG_ROLE_MSG_SIZE);
			if (data == NULL) {
				TRACE_LEAVE2("allocating uba failed");
				return NCSCC_RC_FAILURE;
			}

			ncs_encode_32bit(
			    &data, (uint32_t)mm->rcvr_peer_key.peer_inst_hdl);
			ncs_encode_32bit(
			    &data, mm->info.peer_msg.info.peer_disc.peer_role);
			ncs_enc_claim_space(uba, MBCSV_PEER_CHG_ROLE_MSG_SIZE);

			break;
		}

		default:
			TRACE_LEAVE2("Invalid message type");
			return NCSCC_RC_FAILURE;
		}

		break;
	}
	case MBCSV_EVT_INTERNAL_CLIENT: {
		data = ncs_enc_reserve_space(uba, MBCSV_INT_CLIENT_MSG_SIZE);
		if (data == NULL) {
			TRACE_LEAVE2("allocating uba failed");
			return NCSCC_RC_FAILURE;
		}

		ncs_encode_8bit(
		    &data, mm->info.peer_msg.info.client_msg.type.evt_type);
		ncs_encode_8bit(&data,
				mm->info.peer_msg.info.client_msg.action);
		ncs_encode_32bit(&data,
				 mm->info.peer_msg.info.client_msg.reo_type);
		ncs_encode_32bit(&data,
				 mm->info.peer_msg.info.client_msg.first_rsp);
		ncs_encode_8bit(&data,
				mm->info.peer_msg.info.client_msg.snd_type);
		ncs_encode_32bit(&data,
				 (uint32_t)mm->rcvr_peer_key.peer_inst_hdl);
		ncs_enc_claim_space(uba, MBCSV_INT_CLIENT_MSG_SIZE);

		/* Append user buffer */
		if (mm->info.peer_msg.info.client_msg.type.evt_type !=
		    NCS_MBCSV_MSG_SYNC_SEND_RSP)
			ncs_enc_append_usrbuf(
			    uba, mm->info.peer_msg.info.client_msg.uba.start);
		mds_enc_cb_done = true;

		break;
	}
	default:
		TRACE_LEAVE2("Incorrect message type");
		return NCSCC_RC_FAILURE;
	}

	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
 * Function Name: mbcsv_mds_dec
 *
 * Description  : Decode message received from the MBCSv peer.
 *
 * Arguments     : NCSMDS_CALLBACK_INFO *info: call back info.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.
 *
 * Notes         : None.
 *****************************************************************************/
uint32_t mbcsv_mds_dec(MDS_CLIENT_HDL yr_svc_hdl, NCSCONTEXT *msg,
		       SS_SVC_ID to_svc, NCS_UBAID *uba,
		       MDS_CLIENT_MSG_FORMAT_VER msg_fmat_ver)
{
	uint8_t *data;
	MBCSV_EVT *mm;
	uint8_t data_buff[MBCSV_MAX_SIZE_DATA];

	if (0 == m_NCS_MSG_FORMAT_IS_VALID(msg_fmat_ver,
					   MBCSV_WRT_PEER_SUBPART_VER_MIN,
					   MBCSV_WRT_PEER_SUBPART_VER_MAX,
					   MBCSV_wrt_PEER_msg_fmt_array)) {
		char str[200];
		snprintf(
		    str, sizeof(str),
		    "Msg format version:%d not supported, message from svc-id:%d dropped",
		    msg_fmat_ver, to_svc);
		TRACE_LEAVE2("%s", str);
		return NCSCC_RC_FAILURE;
	}

	if (uba == NULL) {
		TRACE_LEAVE2("userbuf is Null");
		return NCSCC_RC_FAILURE;
	}

	if (msg == NULL) {
		TRACE_LEAVE2("message is Null");
		return NCSCC_RC_FAILURE;
	}

	mm = m_MMGR_ALLOC_MBCSV_EVT;
	if (mm == NULL) {
		TRACE_LEAVE2("malloc failed");
		return NCSCC_RC_FAILURE;
	}

	memset(mm, '\0', sizeof(MBCSV_EVT));

	*msg = mm;

	data = ncs_dec_flatten_space(uba, data_buff, MBCSV_MSG_TYPE_SIZE);
	if (data == NULL) {
		m_MMGR_FREE_MBCSV_EVT(mm);
		TRACE_LEAVE2("decode failed");
		return NCSCC_RC_FAILURE;
	}

	mm->info.peer_msg.type = ncs_decode_8bit(&data);
	mm->rcvr_peer_key.svc_id = ncs_decode_32bit(&data);

	ncs_dec_skip_space(uba, MBCSV_MSG_TYPE_SIZE);

	switch (mm->info.peer_msg.type) {
	case MBCSV_EVT_INTERNAL_PEER_DISC: {
		data =
		    ncs_dec_flatten_space(uba, data_buff, MBCSV_MSG_SUB_TYPE);
		if (data == NULL) {
			m_MMGR_FREE_MBCSV_EVT(mm);
			TRACE_LEAVE2("decode failed");
			return NCSCC_RC_FAILURE;
		}

		mm->info.peer_msg.info.peer_disc.msg_sub_type =
		    ncs_decode_8bit(&data);
		ncs_dec_skip_space(uba, MBCSV_MSG_SUB_TYPE);

		switch (mm->info.peer_msg.info.peer_disc.msg_sub_type) {
		case MBCSV_PEER_UP_MSG: {
			data = ncs_dec_flatten_space(uba, data_buff,
						     MBCSV_PEER_UP_MSG_SIZE);
			if (data == NULL) {
				m_MMGR_FREE_MBCSV_EVT(mm);
				TRACE_LEAVE2("decode failed");
				return NCSCC_RC_FAILURE;
			}
			mm->info.peer_msg.info.peer_disc.peer_role =
			    ncs_decode_32bit(&data);
			ncs_dec_skip_space(uba, MBCSV_PEER_UP_MSG_SIZE);

			mbcsv_decode_version(uba,
					     &mm->info.peer_msg.info.peer_disc
						  .info.peer_up.peer_version);

			break;
		}

		case MBCSV_PEER_DOWN_MSG: {
			data = ncs_dec_flatten_space(uba, data_buff,
						     MBCSV_PEER_DOWN_MSG_SIZE);
			if (data == NULL) {
				m_MMGR_FREE_MBCSV_EVT(mm);
				TRACE_LEAVE2("decode failed");
				return NCSCC_RC_FAILURE;
			}

			mm->rcvr_peer_key.peer_inst_hdl =
			    ncs_decode_32bit(&data);
			mm->info.peer_msg.info.peer_disc.peer_role =
			    ncs_decode_32bit(&data);
			ncs_dec_skip_space(uba, MBCSV_PEER_DOWN_MSG_SIZE);

			break;
		}

		case MBCSV_PEER_INFO_MSG: {
			data = ncs_dec_flatten_space(uba, data_buff,
						     MBCSV_PEER_INFO_MSG_SIZE);
			if (data == NULL) {
				m_MMGR_FREE_MBCSV_EVT(mm);
				TRACE_LEAVE2("decode failed");
				return NCSCC_RC_FAILURE;
			}
			mm->info.peer_msg.info.peer_disc.info.peer_info
			    .my_peer_inst_hdl = ncs_decode_32bit(&data);
			mm->info.peer_msg.info.peer_disc.peer_role =
			    ncs_decode_32bit(&data);
			mm->info.peer_msg.info.peer_disc.info.peer_info
			    .compatible = ncs_decode_8bit(&data);
			ncs_dec_skip_space(uba, MBCSV_PEER_INFO_MSG_SIZE);

			mbcsv_decode_version(uba,
					     &mm->info.peer_msg.info.peer_disc
						  .info.peer_up.peer_version);

			break;
		}

		case MBCSV_PEER_INFO_RSP_MSG: {
			data = ncs_dec_flatten_space(
			    uba, data_buff, MBCSV_PEER_INFO_RSP_MSG_SIZE);
			if (data == NULL) {
				m_MMGR_FREE_MBCSV_EVT(mm);
				TRACE_LEAVE2("decode failed");
				return NCSCC_RC_FAILURE;
			}

			mm->rcvr_peer_key.peer_inst_hdl =
			    ncs_decode_32bit(&data);
			mm->info.peer_msg.info.peer_disc.info.peer_info_rsp
			    .my_peer_inst_hdl = ncs_decode_32bit(&data);
			mm->info.peer_msg.info.peer_disc.peer_role =
			    ncs_decode_32bit(&data);
			mm->info.peer_msg.info.peer_disc.info.peer_info_rsp
			    .compatible = ncs_decode_8bit(&data);
			ncs_dec_skip_space(uba, MBCSV_PEER_INFO_RSP_MSG_SIZE);

			mbcsv_decode_version(
			    uba, &mm->info.peer_msg.info.peer_disc.info
				      .peer_info_rsp.peer_version);

			break;
		}

		case MBCSV_PEER_CHG_ROLE_MSG: {
			data = ncs_dec_flatten_space(
			    uba, data_buff, MBCSV_PEER_CHG_ROLE_MSG_SIZE);
			if (data == NULL) {
				m_MMGR_FREE_MBCSV_EVT(mm);
				TRACE_LEAVE2("decode failed");
				return NCSCC_RC_FAILURE;
			}

			mm->rcvr_peer_key.peer_inst_hdl =
			    ncs_decode_32bit(&data);
			mm->info.peer_msg.info.peer_disc.peer_role =
			    ncs_decode_32bit(&data);
			ncs_dec_skip_space(uba, MBCSV_PEER_CHG_ROLE_MSG_SIZE);

			break;
		}

		default:
			m_MMGR_FREE_MBCSV_EVT(mm);
			TRACE_LEAVE2("incorrect peer sub message type");
			return NCSCC_RC_FAILURE;
		}

		break;
	}
	case MBCSV_EVT_INTERNAL_CLIENT: {
		data = ncs_dec_flatten_space(uba, data_buff,
					     MBCSV_INT_CLIENT_MSG_SIZE);
		if (data == NULL) {
			m_MMGR_FREE_MBCSV_EVT(mm);
			TRACE_LEAVE2("decode failed");
			return NCSCC_RC_FAILURE;
		}

		mm->info.peer_msg.info.client_msg.type.evt_type =
		    ncs_decode_8bit(&data);
		mm->info.peer_msg.info.client_msg.action =
		    ncs_decode_8bit(&data);
		mm->info.peer_msg.info.client_msg.reo_type =
		    ncs_decode_32bit(&data);
		mm->info.peer_msg.info.client_msg.first_rsp =
		    ncs_decode_32bit(&data);
		mm->info.peer_msg.info.client_msg.snd_type =
		    ncs_decode_8bit(&data);
		mm->rcvr_peer_key.peer_inst_hdl = ncs_decode_32bit(&data);
		ncs_dec_skip_space(uba, MBCSV_INT_CLIENT_MSG_SIZE);

		/* Copy user buffer */
		memcpy(&mm->info.peer_msg.info.client_msg.uba, uba,
		       sizeof(NCS_UBAID));

		break;
	}
	default:
		m_MMGR_FREE_MBCSV_EVT(mm);
		TRACE_LEAVE2("incorrect message type");
		return NCSCC_RC_FAILURE;
	}

	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
 * Function Name: mbcsv_mds_cpy
 *
 * Description  : Copy message to be sent to the peer in the same process.
 *
 * Arguments     : NCSMDS_CALLBACK_INFO *info: call back info.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.

 * Notes         : None.
 *****************************************************************************/

uint32_t mbcsv_mds_cpy(MDS_CLIENT_HDL yr_svc_hdl, NCSCONTEXT msg,
		       SS_SVC_ID to_svc, NCSCONTEXT *cpy, bool last,
		       MDS_SVC_PVT_SUB_PART_VER rem_svc_pvt_ver,
		       MDS_CLIENT_MSG_FORMAT_VER *msg_fmt_ver)
{
	MBCSV_EVT *mm;
	MDS_CLIENT_MSG_FORMAT_VER msg_fmt_version;
	TRACE_ENTER();

	if (msg == NULL) {
		TRACE_LEAVE2("msg is NULL");
		return NCSCC_RC_FAILURE;
	}

	/* Set the current message format version */
	msg_fmt_version = m_NCS_ENC_MSG_FMT_GET(
	    rem_svc_pvt_ver, MBCSV_WRT_PEER_SUBPART_VER_MIN,
	    MBCSV_WRT_PEER_SUBPART_VER_MAX, MBCSV_wrt_PEER_msg_fmt_array);
	if (0 == msg_fmt_version) {
		char str[200];
		snprintf(
		    str, sizeof(str),
		    "Peer MDS Subpart version:%d not supported, message to svc-id:%d dropped",
		    rem_svc_pvt_ver, to_svc);
		TRACE_LEAVE2("%s", str);
		return NCSCC_RC_FAILURE;
	}
	*msg_fmt_ver = msg_fmt_version;

	mm = m_MMGR_ALLOC_MBCSV_EVT;

	if (mm == NULL) {
		TRACE_LEAVE2("malloc failed");
		return NCSCC_RC_FAILURE;
	}

	memset(mm, '\0', sizeof(MBCSV_EVT));
	*cpy = mm;

	/*No mem set is require here since we are copying the message */
	memcpy(mm, msg, sizeof(MBCSV_EVT));

	TRACE_LEAVE();
	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
 * Function Name: mbcsv_encode_version
 *
 * Description  : Encode version information in the message to be sent to the
 *                peer.
 *
 * Arguments     : uba  - User Buffer.
 *                 version - Software version.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.
 *
 * Notes         : None.
 *****************************************************************************/

uint32_t mbcsv_encode_version(NCS_UBAID *uba, uint16_t version)
{
	uint8_t *data;
	TRACE_ENTER();

	if (uba == NULL) {
		TRACE_LEAVE2("user buff is NULL");
		return NCSCC_RC_FAILURE;
	}

	data = ncs_enc_reserve_space(uba, MBCSV_MSG_VER_SIZE);

	if (data == NULL) {
		TRACE_LEAVE2("encode failed");
		return NCSCC_RC_FAILURE;
	}

	ncs_encode_16bit(&data, version);

	ncs_enc_claim_space(uba, MBCSV_MSG_VER_SIZE);

	TRACE_LEAVE();
	return NCSCC_RC_SUCCESS;
}

/****************************************************************************
 * Function Name: mbcsv_decode_version
 *
 * Description  : Decode version information from the message received from
 *peer.
 *
 * Arguments     : uba  - User Buffer.
 *                 version - Software version.
 *
 * Return Values : NCSCC_RC_SUCCESS/NCSCC_RC_FAILURE.
 *
 * Notes         : None.
 *****************************************************************************/

uint32_t mbcsv_decode_version(NCS_UBAID *uba, uint16_t *version)
{
	uint8_t *data;
	uint8_t data_buff[MBCSV_MAX_SIZE_DATA];
	TRACE_ENTER();

	if (uba == NULL) {
		TRACE_LEAVE2("userbuf is NULL");
		return NCSCC_RC_FAILURE;
	}

	data = ncs_dec_flatten_space(uba, data_buff, MBCSV_MSG_VER_SIZE);

	if (data == NULL) {
		TRACE_LEAVE2("decode failed");
		return NCSCC_RC_FAILURE;
	}

	*version = ncs_decode_16bit(&data);

	ncs_dec_skip_space(uba, MBCSV_MSG_VER_SIZE);

	TRACE_LEAVE();
	return NCSCC_RC_SUCCESS;
}
